{
  ABOUT

  TRVHTMLViewImporter
  This component imports document from
  THTMLViewer (HTML viewer component originally created by David Baldwin)
  to TRichView.

  TRVHTMLViewImporter
  Copyright (c) 2005 by Andrej Sandorf andrejsandorf@gmx.de
  Copyright (c) 2008-20?? by Sergey Tkachenko svt@trichview.com

  You are free to use, modify and redistribute it with the following restrictions:
  - This header and copyright notice must stay intact
  - You must have a TRichView license http://www.trichview.com/

  ----------------------------------------------------------------------------

  DOWNLOADING THTMLVIEWER

  1) The original version of THTMLViewer is here: www.pbear.com
  It can be installed in Delphi 4..2007.

  2) THTMLViewer was ported to Delphi 2009: www.songbeamer.com/delphi/
  This version can be installed in Delphi 4..XE

  3) THTMLViewer is currently developed at http://code.google.com/p/thtmlviewer/
  This version can be installed in Delphi 4..XE2
  This version is recommended.
  To compile TrvHtmlViewImporter with this version, activate GOOGLECODE define
  below (search for .$DEFINE GOOGLECODE)

  ----------------------------------------------------------------------------

  BEFORE INSTALLING (complete information is in ReadMe.txt)

  -------- FOR pbear.com AND songbeamer.com VERSIONS: -----

  In HTMLView there must be a small source change to access
  variables which are declared as private.

  In "Htmlsubs.pas" in class "TBlockLI"
   TBlockLI = class(TBlock)
    private
    protected // <-- insert this, so we can access ListType, etc.
      ListType: ListTypeType;
      ListNumb: integer;
    ...

  In "HTMLUn2.pas" in class "TChPosObj"
  TChPosObj = class (TIDObject)
  public < -- change "private" to "public", so we can access ChPos
    ChPos: integer;
    List: TList;
   ...


  ----------------------------------------------------------------------------

  BRIEF MANUAL

  Properties:

  - AutoHyperlinks: Boolean (default: True)
    If True, hypertext targets are assigned directly to item tags.
    If False, TRichView.OnReadHyperlink event occurs instead.

  - ImgSrcToImageName: Boolean (default: False)
    If True, <img src> is assigned to items names. It is not necessary, because
    it is always assigned to rvespImageFileName items properties.

  - HTMLViewer: THTMLViewer (default: nil)
    If assigned, you can call ImportHtmlViewer method without htmlview parameter.

  - DefaultHMargin: Integer (default: 5)
    Margin value used for LeftMargin and RightMargin, if not specified in HTML.
  - DefaultVMargin: Integer (default: 5)
    Margin value used for TopMargin and BottomMargin, if not specified in HTML.

  - CompleteLists: Boolean (default: True)
    Adds levels to lists loaded from HTML, up to 9 levels
    (to make them similar to MS Word's lists)

  - UseCommonBulletLists: Boolean (default: True)
    This property is ignored

  Methods:

  procedure ImportHtmlViewer(htmlview: THTMLViewer; rve: TCustomRichView;
    const BasePath: String='');
  procedure ImportHtmlViewer(rve: TCustomRichView; const BasePath: String='');
  These methods load HTML from HTML viewer to TRichView.
  After these methods, call NormalizeRichView(rve.RVData)
  (defined in RVNormalize.pas, in RichViewActions) and rve.Format.
  BasePath is the base path for images (if not specified, the current application
  directory is used)

  procedure AppendHtmlViewer(htmlview: THTMLViewer; rve: TCustomRichView'
    const BasePath: String='');
  The same as above, but appends the loaded document to the end, without clearing
  the existing document. This method does not read background and margins.

  function LoadFromClipboard(rve: TCustomRichView;
    htmlview: THTMLViewer = nil): Boolean;
  Loads HTML from the Clipboard (if available)

}

// Activate this define if you use THMLViewer from http://code.google.com/p/thtmlviewer/

{$DEFINE GOOGLECODE}

unit rvHtmlViewImport;

interface

{$I RV_Defs.inc}

uses
  Windows, Classes, Graphics, Forms,
  {$IFDEF RICHVIEWDEF6}
  Variants,
  {$ENDIF}
  {$IFDEF RICHVIEWDEFXE2}
  UITypes,
  {$ENDIF}

  //HtmlView Component
  Htmlview, ReadHTML, HTMLsbs1, Htmlsubs, HTMLun2, StyleUn,
  {$IFDEF GOOGLECODE}
  StyleTypes, HtmlFonts, HtmlImages,
  {$ENDIF}

  //RichView Component
  RichView, RVStyle, RVClasses, RVScroll, RVItem, RVFuncs,
  RVTable, CRVData, CRVFData, RVUni, RVTypes;


type
  TRVCurrentListRecord = record
    ListNo, Level: Integer;
    Counter: Integer;
    AssignedLevels: set of Byte;
  end;


  TRVHTMLViewImporter = class (TComponent)
  private
    FRVE: TCustomRichView;
    FRVS: TRVStyle;
    FUseCommonBulletLists, FCompleteLists: Boolean;
    FListCount: Integer;
    FAutoHyperlinks: Boolean;
    FImgSrcToImageName: Boolean;
    FHTMLViewer: THTMLViewer;
    FDefaultHMargin: Integer;
    FDefaultVMargin: Integer;
    FAnchorList: TRVList;
    FTargetBlockList: TRVList;
    FAnchorIndex: Integer;
    FBasePath: String;
    {
    FDefFontSize: Integer;
    FDefLinkColor: TColor;
    FDefFontColor: TColor;
    FDefFontName: TFontName;
    FDefPreFontName: TFontName;
    FDefColor: TColor;
    FUndelineLinks: Boolean;
    }
    procedure ImportBlock(RVData: TCustomRVData; block: TBlock;
      DefParaNo: Integer; var ListRec: TRVCurrentListRecord; Indent: TRVStyleLength);
    procedure ImportBlockLi(RVData: TCustomRVData; blockli: TBlockLI;
      ParaNo: Integer; var ListRec: TRVCurrentListRecord; Indent: TRVStyleLength);
    procedure ImportListBlock(RVData: TCustomRVData; block: TListBlock;
      DefParaNo: Integer; var ListRec: TRVCurrentListRecord; Indent: TRVStyleLength);
    procedure ImportHorzLine(RVData: TCustomRVData; blockhl: THorzLine;
      ParaNo: Integer);
    procedure ImportHRBlock(RVData: TCustomRVData; blockhr: THRBlock;
      DefParaNo: Integer);
    procedure ImportPicture(RVData: TCustomRVData; ImageObj: TImageObj;
      FO: TFontObj; const ParaNo: Integer);
    procedure ImportSection(RVData: TCustomRVData; sec: TSection;
      DefParaNo: Integer; FirstSection: Boolean);
    procedure ImportTable(RVData: TCustomRVData; htmltable: ThtmlTable;
      parentBlock: TBlock; ParaNo: integer);
    procedure ImportTableAndCaptionBlock(RVData: TCustomRVData;
      tableandcaption: TTableAndCaptionBlock; ParaNo: Integer);
    procedure ImportTableBlock(RVData: TCustomRVData; tableblock: TTableBlock;
      ParaNo: Integer);
    procedure ImportBody(RVData: TCustomRVData; bodyblock: TBodyBlock);
    procedure ImportTableCell(tablerow: TRVTableRow; cellobj: TCellObj;
      colindex: integer; var SumHPadding, SumVPadding,
      PaddingHCount, PaddingVCount: Integer; IsTableBorder: Boolean);
    procedure ImportTableRow(table: TRVTableItemInfo; row: TCellList;
      rowindex: integer; var SumHPadding, SumVPadding,
      PaddingHCount, PaddingVCount, CellCount: Integer;
      IsTableBorder: Boolean);
    procedure DoImport(SectionList: {$IFDEF GOOGLECODE}ThtDocument{$ELSE}TSectionList{$ENDIF};
      rve: TCustomRichView;
      AppendMode: Boolean);
    procedure MakeTag(const URL: String; var StyleNo: Integer; var Tag: TRVTag;
      var ItemName: TRVRawByteString);
    procedure LoadClipboardHTML(const Data: TRVRawByteString; rve: TCustomRichView;
      htmlview: THTMLViewer);
    function GetFullFileName(const FileName: String): String;
    procedure AdjustBasePath(EndingSlash: Boolean);
    procedure DoCompleteList(ListNo: Integer);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure ImportHtmlViewer(htmlview: THTMLViewer;
      rve: TCustomRichView; const BasePath: String=''); overload;
    procedure ImportHtmlViewer(rve: TCustomRichView; const BasePath: String=''); overload;
    procedure AppendHtmlViewer(htmlview: THTMLViewer;
      rve: TCustomRichView; const BasePath: String='');
    function LoadFromClipboard(rve: TCustomRichView;
      htmlview: THTMLViewer = nil): Boolean;
    {
    procedure LoadFromString(const s: String; rve: TCustomRichView);
    procedure LoadFromStream(Stream: TStream; rve: TCustomRichView);
    function LoadFromFile(const FileName: String; rve: TCustomRichView): Boolean;
    property DefFontName: TFontName read FDefFontName write FDefFontName;
    property DefPreFontName: TFontName read FDefPreFontName write FDefPreFontName;
    property DefFontSize: Integer read FDefFontSize write FDefFontSize default 12;
    property DefFontColor: TColor read FDefFontColor write FDefFontColor
      default clWindowText;
    property DefLinkColor: TColor read FDefLinkColor write FDefLinkColor
      default clBlue;
    property DefColor: TColor read FDefColor write FDefColor default clWindow;
    property UndelineLinks: Boolean read FUndelineLinks write FUndelineLinks
      default False;
    }
  published
    property UseCommonBulletLists: Boolean read FUseCommonBulletLists
      write FUseCommonBulletLists default True; // ignored
    property CompleteLists: Boolean read FCompleteLists write FCompleteLists
      default True;
    property AutoHyperlinks: Boolean read FAutoHyperlinks write FAutoHyperlinks
      default True;
    property ImgSrcToImageName: Boolean read FImgSrcToImageName write FImgSrcToImageName
      default False;
    property HTMLViewer: THTMLViewer read FHTMLViewer write FHTMLViewer;
    property DefaultHMargin: Integer read FDefaultHMargin write FDefaultHMargin default 5;
    property DefaultVMargin: Integer read FDefaultVMargin write FDefaultVMargin default 5;
  end;

  procedure Register;

implementation
uses
{$IFDEF DEBUG}
  Dialogs,
{$ENDIF DEBUG}
  SysUtils, RVStr, RVGrHandler;

type
  THackFloatingObj = class (TFloatingObj);
  THackHorzLine = class (THorzLine);
  THackBlockLi = class (TBlockLI);
  THackChPosObj = class(TChPosObj);

const
  ERR_NOT_DEF = 'HTMLViewer is not defined';

procedure LogNotImplemented(const s: string);
begin
{$IFDEF DEBUG}
  ShowMessage(s);
{$ENDIF DEBUG}
end;
{------------------------------------------------------------------------------}
{ Utility functions }
{------------------------------------------------------------------------------}
procedure InitListRec(var ListRec: TRVCurrentListRecord);
begin
  ListRec.ListNo := -1;
  ListRec.Level := -1;
  ListRec.Counter := 0;
  ListRec.AssignedLevels := [];
end;
{------------------------------------------------------------------------------}
procedure LoadPicture(RVData: TCustomRVData; Source: String;
  var gr: TGraphic; w,h: Integer; UseInvalidPicture: Boolean);
var Invalid: Boolean;
    DecodedSource: String;
begin
  Invalid := False;
  gr := RVData.ImportPicture(Source, w, h, Invalid);
  if gr=nil then begin
    Source := RV_DecodeFileURL(Source);
    gr := RVGraphicHandler.LoadFromFile(Source);
  end;
  if gr=nil then begin
    DecodedSource := RV_DecodeURL(Source, False);
    if DecodedSource<>Source then begin
      Source := DecodedSource;
      gr := RVGraphicHandler.LoadFromFile(Source);
    end;
  end;
  if (gr=nil) and UseInvalidPicture then begin
    gr := RVGraphicHandler.CreateGraphic(
      TGraphicClass(RVData.GetRVStyle.InvalidPicture.Graphic.ClassType));
    gr.Assign(RVData.GetRVStyle.InvalidPicture.Graphic);
  end;
end;
{------------------------------------------------------------------------------}
procedure RemoveExtraTables(rve: TCustomRichView);
var table: TRVTableItemInfo;
    RVData: TCustomRVData;
    RVFOptions: TRVFOptions;
    Stream: TRVMemoryStream;
begin
  RVData := rve.RVData;
  while True do begin
    if (RVData.ItemCount<>1) or (RVData.GetItemStyle(0)<>rvsTable) then
      break;
    table := TRVTableItemInfo(RVData.GetItem(0));
    if (table.RowCount<>1) or (table.ColCount<>1) then
      break;
    RVData := table.Cells[0,0].GetRVData;
  end;
  if RVData=rve.RVData then
    exit;
  RVFOptions := rve.RVFOptions;
  rve.RVFOptions := RVFOptions-[rvfoSaveTextStyles, rvfoSaveParaStyles];
  try
    Stream := TRVMemoryStream.Create;
    try
      RVData.SaveRVFToStream(Stream, False, clNone, nil, nil);
      rve.Clear;
      Stream.Position := 0;
      rve.LoadRVFFromStream(Stream);
    finally
      Stream.Free;
    end;
  finally
    rve.RVFOptions := RVFOptions;
  end;
end;
{------------------------------------------------------------------------------}
type
{$IFDEF GOOGLECODE}
  TRVHtmlBorderStyle = ThtBorderStyle;
  TRVHtmlMarginArray = ThtMarginArray;
  TRVHtmlJustify = ThtJustify;
{$ELSE}
  TRVHtmlBorderStyle = BorderStyleType;
  TRVHtmlMarginArray = TMarginArray;
  TRVHtmlJustify = JustifyType;
{$ENDIF}

procedure FillVisibleBorders(const MargArray: TRVHtmlMarginArray;
  VisibleBorders: TRVBooleanRect);
begin
  VisibleBorders.Left := TRVHtmlBorderStyle(MargArray[BorderLeftStyle])<>bssNone;
  VisibleBorders.Right := TRVHtmlBorderStyle(MargArray[BorderRightStyle])<>bssNone;
  VisibleBorders.Top := TRVHtmlBorderStyle(MargArray[BorderTopStyle])<>bssNone;
  VisibleBorders.Bottom := TRVHtmlBorderStyle(MargArray[BorderBottomStyle])<>bssNone;
end;
{------------------------------------------------------------------------------}
function GetBlockBorderWidth(const MargArray: TRVHtmlMarginArray): Integer;
begin
  Result := 0;
  if TRVHtmlBorderStyle(MargArray[BorderLeftStyle])<>bssNone then
    Result := MargArray[BorderLeftWidth];
  if (TRVHtmlBorderStyle(MargArray[BorderRightStyle])<>bssNone) and
    (MargArray[BorderRightWidth]>Result) then
    Result := MargArray[BorderRightWidth];
  if (TRVHtmlBorderStyle(MargArray[BorderTopStyle])<>bssNone) and
    (MargArray[BorderTopWidth]>Result) then
    Result := MargArray[BorderTopWidth];
  if (TRVHtmlBorderStyle(MargArray[BorderBottomStyle])<>bssNone) and
    (MargArray[BorderBottomWidth]>Result) then
    Result := MargArray[BorderBottomWidth];
end;
{------------------------------------------------------------------------------}
function GetBlockBorderColor(const MargArray: TRVHtmlMarginArray): TColor;
begin
  Result := clBlack;
  if TRVHtmlBorderStyle(MargArray[BorderLeftStyle])<>bssNone then
    Result := MargArray[BorderLeftColor]
  else if TRVHtmlBorderStyle(MargArray[BorderTopStyle])<>bssNone then
    Result := MargArray[BorderTopColor]
  else if TRVHtmlBorderStyle(MargArray[BorderRightStyle])<>bssNone then
    Result := MargArray[BorderRightColor]
  else if TRVHtmlBorderStyle(MargArray[BorderBottomStyle])<>bssNone then
    Result := MargArray[BorderBottomColor];
end;
{------------------------------------------------------------------------------}
function GetBlockBorderStyle(block: TBlock): TRVHtmlBorderStyle;
begin
  {$IFnDEF GOOGLECODE}
  Result := block.BorderStyle;
  {$ELSE}
  Result := bssNone;
  if not block.HasBorderStyle then
    exit;
  if TRVHtmlBorderStyle(block.MargArray[BorderLeftStyle])<>bssNone then
    Result := TRVHtmlBorderStyle(block.MargArray[BorderLeftStyle])
  else if TRVHtmlBorderStyle(block.MargArray[BorderTopStyle])<>bssNone then
    Result := TRVHtmlBorderStyle(block.MargArray[BorderTopStyle])
  else if TRVHtmlBorderStyle(block.MargArray[BorderRightStyle])<>bssNone then
    Result := TRVHtmlBorderStyle(block.MargArray[BorderRightStyle])
  else if TRVHtmlBorderStyle(block.MargArray[BorderBottomStyle])<>bssNone then
    Result := TRVHtmlBorderStyle(block.MargArray[BorderBottomStyle])
  {$ENDIF}
end;
{------------------------------------------------------------------------------}
function GetBlockBorderLightColor(const MargArray: TRVHtmlMarginArray): TColor;
begin
  case TRVHtmlBorderStyle(MargArray[BorderLeftStyle]) of
    bssNone:
        case TRVHtmlBorderStyle(MargArray[BorderTopStyle]) of
          bssNone, bssInset:
            Result := clBtnHighLight;
          bssOutset:
            Result := clBtnShadow;
         else
            Result := MargArray[BorderTopColor];
        end;
    bssInset:
      Result := clBtnHighLight;
    bssOutset:
      Result := clBtnShadow;
    else
      Result := MargArray[BorderLeftColor]
  end;
end;
{------------------------------------------------------------------------------}
function GetBlockBorderDarkColor(const MargArray: TRVHtmlMarginArray): TColor;
begin
  case TRVHtmlBorderStyle(MargArray[BorderRightStyle]) of
    bssNone:
        case TRVHtmlBorderStyle(MargArray[BorderBottomStyle]) of
          bssNone, bssInset:
            Result := clBtnShadow;
          bssOutset:
            Result := clBtnHighLight;
         else
            Result := MargArray[BorderBottomColor];
        end;
    bssInset:
      Result := clBtnShadow;
    bssOutset:
      Result := clBtnHighLight;
    else
      Result := MargArray[BorderRightColor]
  end;
end;
{------------------------------------------------------------------------------}
function GetRVAlignment(Justify: TRVHtmlJustify): TRVAlignment;
begin
  case Justify of
    Centered:    Result := rvaCenter;
    Right:       Result := rvaRight;
    FullJustify: Result := rvaJustify;
    else         Result := rvaLeft;
  end;
end;
{------------------------------------------------------------------------------}
function GetSpacing(obj: TFloatingObj): Integer;
begin
  Result := THackFloatingObj(obj).HSpaceL;
  if Result>THackFloatingObj(obj).HSpaceR then
    Result := THackFloatingObj(obj).HSpaceR;
  if Result>THackFloatingObj(obj).VSpaceT then
    Result := THackFloatingObj(obj).VSpaceT;
  if Result>THackFloatingObj(obj).VSpaceB then
    Result := THackFloatingObj(obj).VSpaceB;
end;
{------------------------------------------------------------------------------}
function GetCellBorderWidth(cellobj: TCellObj): Integer;
begin
  Result := 0;
  if TRVHtmlBorderStyle(cellobj.MargArray[BorderLeftStyle])<>bssNone then
    Result := cellobj.BrdLeft;
  if (TRVHtmlBorderStyle(cellobj.MargArray[BorderRightStyle])<>bssNone) and
    (cellobj.BrdRight>Result) then
    Result := cellobj.BrdRight;
  if (TRVHtmlBorderStyle(cellobj.MargArray[BorderTopStyle])<>bssNone) and
    (cellobj.BrdTop>Result) then
    Result := cellobj.BrdTop;
  if (TRVHtmlBorderStyle(cellobj.MargArray[BorderBottomStyle])<>bssNone) and
    (cellobj.BrdBottom>Result) then
    Result := cellobj.BrdBottom;
end;
{------------------------------------------------------------------------------}
procedure AddCellHPadding(cellobj: TCellObj; var CellPadding, Count: Integer);
begin
  if cellobj.MargArray[PaddingLeft]>=0 then begin
    inc(CellPadding, cellobj.MargArray[PaddingLeft]);
    inc(Count);
  end;
  if cellobj.MargArray[PaddingRight]>=0 then begin
    inc(CellPadding, cellobj.MargArray[PaddingRight]);
    inc(Count);
  end;
end;
{------------------------------------------------------------------------------}
procedure AddCellVPadding(cellobj: TCellObj; var CellPadding, Count: Integer);
begin
  if cellobj.MargArray[PaddingTop]>=0 then begin
    inc(CellPadding, cellobj.MargArray[PaddingTop]);
    inc(Count);
  end;
  if cellobj.MargArray[PaddingBottom]>=0 then begin
    inc(CellPadding, cellobj.MargArray[PaddingBottom]);
    inc(Count);
  end;
end;
{------------------------------------------------------------------------------}
function ExtractDomain(const URL: String): String;
var p,p2: Integer;
    Prefix: String;
begin
  Result := URL;
  p := Pos('://', Result);
  if p>0 then begin
    Prefix := Copy(Result, 1, p+2);
    Result := Copy(Result, p+3, Length(Result));
    end
  else
    Prefix := '';
  p := Pos('/', Result);
  p2 := Pos('\', Result);
  if (p2>0) and (p2<p) then
    p := p2;
  p2 := Pos('?', Result);
  if (p2>0) and (p2<p) then
    p := p2;
  if p>0 then
    Result := Copy(Result, 1, p-1);
  Result := Prefix+Result;
end;
{------------------------------------------------------------------------------}
// Copied from StyleUn with minor modifications
{$HINTS OFF}
function LengthConv(const Str: string; Relative: boolean; Base, EmSize, ExSize,
                          Default: integer): integer;
{given a length string, return the appropriate pixel value.  Base is the
 base value for percentage. EmSize, ExSize for units relative to the font.
 Relative makes a numerical entry relative to Base.
 Default returned if no match.}
var
  V, PPI: double;
  S1, S2: string;
  I, J: integer;
begin
  if Str='auto' then begin
    Result := Default;
    exit;
  end;
  PPI := RichViewPixelsPerInch;
  if PPI=0 then
    PPI := Screen.PixelsPerInch;
  Val(Str, V, I);
  J := Pos('e', Str);   {'e' would be legal for Val but not for us}
  if (J <> 0) and (I > J) then
  I := J;
  if I > 0 then
    begin
    S1 := Copy(Str, 1, I-1);
    S2 := Trim(Copy(Str, I, Length(Str)-I+1));
    end
  else
    begin
    S1 := Str;
    S2 := '';
    end;
  {S1 has the number, S2 the units}
  Val(S1, V, I);
  if S2 = 'in' then
    V := PPI*V
  else if S2 = 'cm' then
    V := PPI*V/2.54
  else if S2 = 'mm' then
    V := PPI*V/25.4
  else if S2 = 'pt' then
    V := V*PPI/72.0
  else if S2 = 'px' then
  else if S2 = 'pc' then
    V := V*PPI/6.0
  else if S2 = 'em' then
    V := V*EmSize
  else if S2 = 'ex' then
    V := V*ExSize
  else if S2 = '%' then
    V := V*Base/100.0
  else if S2 = '' then
    if Relative then
      V := V*Base      {relative}
    else
      V := V           {same as pixels, at least for margins}
  else V := Default;
  Result := Round(V);
end;
{$HINTS ON}
{------------------------------------------------------------------------------}
function RemoveLineBreaks(const s: String): String;
begin
  Result := StringReplace(AdjustLineBreaks(s), #13#10, ' ', [rfReplaceAll]);
end;
{------------------------------------------------------------------------------}
procedure DoMerge(rvt: TRVTableItemInfo; CurrRow, CurrCol, ColSpan, RowSpan: integer);
var
  i, j, colcount: integer;
begin
  if ColSpan < 1 then
    ColSpan := 1;
  if RowSpan < 1 then
    RowSpan := 1;
  if (ColSpan > 1) or (RowSpan > 1) then
  begin

    while CurrRow + RowSpan > RVt.Rows.Count do
      RVt.Rows.Add(Rvt.Rows[0].Count);

    colcount := CurrCol + ColSpan - RVT.Rows[0].Count;

    for i := 0 to RVT.Rows.Count - 1 do
      for j := 0 to colcount - 1 do
        RVT.Rows[i].Add;

    RVT.MergeCells(CurrRow, CurrCol, ColSpan, rowSpan, True);
  end;
end;
{======================= Classes for finding anchors ==========================}
type
  TDocPosItem = class
    public
      Pos: Integer;
      ID: String;
  end;
  TBlockIDItem = class
    public
      Block: Pointer;
      ID: String;
  end;


function DocPosSortCompare(Item1, Item2: Pointer): Integer;
begin
  Result := TDocPosItem(Item1).Pos-TDocPosItem(Item2).Pos;
end;

function GetBlockID(Block: Pointer; List: TList): String;
var i: Integer;
begin
  Result := '';
  for i := 0 to List.Count-1 do
    if TBlockIDItem(List.Items[i]).Block=Block then begin
      Result := TBlockIDItem(List.Items[i]).ID;
      exit;
    end;
end;

{============================ TRVHTMLViewImporter =============================}
procedure TRVHTMLViewImporter.ImportHtmlViewer(htmlview: THTMLViewer;
  rve: TCustomRichView; const BasePath: String='');
begin
  if Htmlview=nil then
    Htmlview := FHTMLViewer;
  FBasePath := BasePath;
  AdjustBasePath(True);
  try
    DoImport(Htmlview.SectionList, rve, False);
  finally
    FBasePath := '';
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVHTMLViewImporter.ImportHtmlViewer(rve: TCustomRichView;
  const BasePath: String='');
begin
  if FHTMLViewer=nil then
    raise Exception.Create(ERR_NOT_DEF);
  FBasePath := BasePath;
  AdjustBasePath(True);
  try
    DoImport(FHTMLViewer.SectionList, rve, False);
  finally
    FBasePath := '';
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVHTMLViewImporter.AppendHtmlViewer(htmlview: THTMLViewer;
  rve: TCustomRichView; const BasePath: String='');
begin
  if Htmlview=nil then
    Htmlview := FHTMLViewer;
  if Htmlview=nil then
    raise Exception.Create(ERR_NOT_DEF);
  FBasePath := BasePath;
  AdjustBasePath(True);
  try
    DoImport(Htmlview.SectionList, rve, True);
  finally
    FBasePath := '';
  end;
end;
{------------------------------------------------------------------------------}
function TRVHTMLViewImporter.GetFullFileName(const FileName: String): String;
var i: Integer;
begin
  if ((Pos(':', FileName)>0) and
     (RVIsURL(FileName) or ((Length(FileName)>1) and (FileName[2]=':')))) or
     ((FileName<>'') and (FileName[1]='#')) then
    Result := FileName
  else if (FileName<>'') and ((FileName[1]='\') or (FileName[1]='/')) then
    Result := ExtractDomain(FBasePath)+FileName
  else
    Result := FBasePath+FileName;
  if Pos('/',FBasePath)<>0 then
    for i := 1 to Length(Result) do
      if Result[i]='\' then
        Result[i] := '/';
end;
{------------------------------------------------------------------------------}
function TRVHTMLViewImporter.LoadFromClipboard(rve: TCustomRichView;
  htmlview: THTMLViewer = nil): Boolean;
var
  cf: cardinal;
  Data: TRVAnsiString;
  mem: cardinal;
  ptr: Pointer;
begin
  Result := False;
  cf := RegisterClipboardFormat('HTML Format');
  if not IsClipboardFormatAvailable(cf) then
    exit;
  OpenClipboard(0);
  try
    mem := GetClipboardData(cf);
    SetLength(Data, GlobalSize(mem));
    ptr := GlobalLock(mem);
    Move(ptr^, PRVAnsiChar(Data)^, Length(Data));
    GlobalUnlock(mem);
  finally
    CloseClipboard;
  end;
  LoadClipboardHtml(Data, rve, htmlview);
  Result := True;
end;
{------------------------------------------------------------------------------}
function ConvertUTF8ToString(const s: TRVRawByteString): String;
begin
  {$IFDEF RVUNICODESTR}
  Result := UTF8ToString(s)
  {$ELSE}
  Result := UTF8Decode(s)
  {$ENDIF}
end;
{------------------------------------------------------------------------------}
procedure TRVHTMLViewImporter.AdjustBasePath(EndingSlash: Boolean);
begin
  if Pos('file://', FBasePath) = 1 then
    FBasePath := Copy(FBasePath, 8,Length(FBasePath));
  if (Pos(':', FBasePath)<>0) and (FBasePath[1]='/') then
    Delete(FBasePath, 1, 1);
  if EndingSlash and (FBasePath<>'') and (FBasePath[Length(FBasePath)]<>'\') and
    (FBasePath[Length(FBasePath)]<>'/') then
    FBasePath := FBasePath+'/';
end;
{------------------------------------------------------------------------------}
procedure TRVHTMLViewImporter.LoadClipboardHTML(const Data: TRVRawByteString;
  rve: TCustomRichView; htmlview: THTMLViewer);
const
  StrSignature = 'Version:';
  StrStartHTML = 'StartHTML:';
  StrEndHTML = 'EndHTML:';
  StrStartFragment = 'StartFragment:';
  StrEndFragment = 'EndFragment:';
  StrSourceURL = 'SourceURL:';
  //StrStartSelection = 'StartSelection:';
  //StrEndSelection = 'EndSelection';
var
  p1, p2: integer;
  s: TRVRawByteString;
  {..........................................................}
  function ReadString(const Key: TRVRawByteString): TRVRawByteString;
  var
    p: integer;
  begin
    Result := '';
    p := RVPos(Key, Data);
    if p = 0 then
      exit;
    inc(p, Length(Key));
    Result := '';
    while not (Data[p] in [#10, #13]) do
    begin
      Result := Result + Data[p];
      inc(p);
    end;
  end;
  {..........................................................}
  function ReadInteger(const Key: TRVRawByteString): integer;
  var
    p: integer;
    val: TRVRawByteString;
  begin
    p := RVPos(Key, Data);
    if p = 0 then
      abort;
    inc(p, Length(Key));
    val := '';
    while Data[p] in ['0'..'9'] do
    begin
      val := val + Data[p];
      inc(p);
    end;
    Result := RVStrToInt(Val);
  end;
  {..........................................................}
  function ExtractPath(const S: string): String;
  var l,i: Integer;
      NotFinish: Boolean;
  begin
    l := Length(S);
    while l > 0 do begin
      if ((S[l]='\') or (S[l]='/') or (S[l]=':')) and (ByteType(S, l)<>mbTrailByte) then begin
        NotFinish := False;
        if S[l]=':' then begin
          for i := 1 to l-1 do
            if ((S[l]='\') or (S[l]='/') or (S[l]=':')) and (ByteType(S, l)<>mbTrailByte) then begin
              NotFinish := True;
              break;
            end;
        end;
        if not NotFinish then
          break;
      end;
      dec(l);
    end;
    if l<0 then
      l := 0;
    Result := Copy(S, 1, l);
  end;
  {..........................................................}
begin
  if Copy(Data, 1, Length(StrSignature)) <> StrSignature then
    abort;
  p1 := ReadInteger(StrStartHTML);
  if p1 < 0 then
    p1 := ReadInteger(StrStartFragment);
  p2 := ReadInteger(StrEndHTML);
  if p2 < 0 then
    p2 := ReadInteger(StrEndFragment);
  FBasePath := ConvertUTF8ToString(ReadString(StrSourceURL));
  AdjustBasePath(False);
  FBasePath := ExtractPath(FBasePath);
  s := #$EF+#$BB+#$BF+Copy(Data, p1, p2 - p1);
  if htmlview = nil then

    htmlview := HTMLViewer;
  try
    {$IFDEF GOOGLECODE}
    htmlview.ServerRoot := 'Z:\';
    {$IFDEF RVUNICODESTR}
    htmlview.LoadFromString(UTF8ToString(s), '');
    {$ELSE}
    htmlview.LoadFromString(UTF8Decode(s), '');
    {$ENDIF}
    htmlview.ServerRoot := '';
    {$ELSE}
    htmlview.LoadFromString(s, 'Z:\');
    {$ENDIF}
    DoImport(htmlview.SectionList, rve, False);
    RemoveExtraTables(rve);
  finally
    FBasePath := '';
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVHTMLViewImporter.MakeTag(const URL: String; var StyleNo: Integer;
  var Tag: TRVTag; var ItemName: TRVRawByteString);
begin
  if StyleNo=rvsPicture then
    StyleNo := rvsHotPicture;
  if AutoHyperlinks then begin
    {$IFDEF RVOLDTAGS}
    if rvoTagsArePChars in FRVE.Options then
      Tag := Integer(StrNew(PChar(URL)));
    {$ELSE}
    Tag := URL;
    {$ENDIF}
    end
  else begin
    Tag := RVEMPTYTAG;
    FRVE.RVData.ReadHyperlink(URL, '', rvlfHTML, StyleNo, Tag, ItemName);
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVHTMLViewImporter.ImportBlock(RVData: TCustomRVData;
  block: TBlock; DefParaNo: Integer; var ListRec: TRVCurrentListRecord;
  Indent: TRVStyleLength);


  function Convert(V: Variant; block: TBlock): integer;
  begin
    if {$IFDEF RICHVIEWDEF6}VarIsStr(V){$ELSE}VarType(V) = VarString{$ENDIF} then
      Result := LengthConv(V, False, 400, block.EmSize, block.ExSize, 0)
    else
      if VarType(V) in varInt then begin
        if V = IntNull then
          Result := 0
        else
          if V = AutoParagraph then
            Result := ParagraphSpace div 2
          else
            Result := V;
        end
      else
        Result := 0;
  end;

  function DoParaFromBlock: integer;
  var ParaInfo: TParaInfo;
      HTMLBorderStyle: TRVHtmlBorderStyle;
  begin
    {if (Block is TTableBlock) and (DefParaNo>=0) then begin
      Result := DefParaNo;
      exit;
    end;}
    ParaInfo := TParaInfo.Create(nil);
    if DefParaNo>=0 then
      ParaInfo.Assign(FRVS.ParaStyles.Items[DefParaNo]);

    if not (block is TBodyBlock) then
    begin
      ParaInfo.SpaceBefore := FRVS.PixelsToUnits(Convert(block.MargArrayO[MarginTop], block)+block.MargArray[PaddingTop]);
      ParaInfo.SpaceAfter := FRVS.PixelsToUnits(Convert(block.MargArrayO[MarginBottom], block)+block.MargArray[PaddingBottom]);
      ParaInfo.LeftIndent := FRVS.PixelsToUnits(Convert(block.MargArrayO[MarginLeft], block)+block.MargArray[PaddingLeft]);
      ParaInfo.RightIndent := FRVS.PixelsToUnits(Convert(block.MargArrayO[MarginRight], block)+block.MargArray[PaddingRight]);
    end;

    if (Length(Block.TagClass)>=3) and (Block.TagClass[1]='h') and
      (Block.TagClass[2]>='1') and (Block.TagClass[2]<='6') then
      ParaInfo.OutlineLevel := ord(Block.TagClass[2])-ord('0');

    HTMLBorderStyle := GetBlockBorderStyle(block);
    if (HTMLBorderStyle<>bssNone) and not (Block is TTableBlock) then begin
      case HTMLBorderStyle of
        bssDouble:
          begin
            ParaInfo.Border.Style := rvbDouble;
            ParaInfo.Border.InternalWidth := FRVS.PixelsToUnits(GetBlockBorderWidth(block.MargArray)-2);
          end
        else
          begin
            ParaInfo.Border.Style := rvbSingle;
            ParaInfo.Border.Width := FRVS.PixelsToUnits(GetBlockBorderWidth(block.MargArray));
          end;
      end;
      ParaInfo.Border.Color := GetBlockBorderColor(block.MargArray);
      ParaInfo.Border.BorderOffsets.Top := FRVS.PixelsToUnits(block.MargArray[PaddingTop]);
      ParaInfo.Border.BorderOffsets.Bottom := FRVS.PixelsToUnits(block.MargArray[PaddingBottom]);
      ParaInfo.Border.BorderOffsets.Left := FRVS.PixelsToUnits(block.MargArray[PaddingLeft]);
      ParaInfo.Border.BorderOffsets.Right := FRVS.PixelsToUnits(block.MargArray[PaddingRight]);
      FillVisibleBorders(block.MargArray, ParaInfo.Border.VisibleBorders);
    end;

    if block.KeepIntact then
      ParaInfo.Options := ParaInfo.Options+[rvpaoKeepLinesTogether];

    if block is TTableAndCaptionBlock then
      ParaInfo.Alignment := GetRVAlignment(TTableAndCaptionBlock(block).Justify)
    else if block is TTableBlock then
      ParaInfo.Alignment := GetRVAlignment(TTableBlock(block).Justify)
    else if (block.MargArray[BackgroundColor] <> clNone) and not (Block is TBodyBlock) then begin
        ParaInfo.Background.Color := block.MargArray[BackgroundColor];
        ParaInfo.Background.BorderOffsets.Top := FRVS.PixelsToUnits(block.MargArray[PaddingTop]);
        ParaInfo.Background.BorderOffsets.Bottom := FRVS.PixelsToUnits(block.MargArray[PaddingBottom]);
        ParaInfo.Background.BorderOffsets.Left := FRVS.PixelsToUnits(block.MargArray[PaddingLeft]);
        ParaInfo.Background.BorderOffsets.Right := FRVS.PixelsToUnits(block.MargArray[PaddingRight]);
    end;
    Result :=
      FRVS.ParaStyles.FindSuchStyle(0, ParaInfo, RVAllParaInfoProperties);
    if Result < 0 then
    begin
      result := FRVS.ParaStyles.Add.Index;
      FRVS.ParaStyles[Result].Assign(ParaInfo);
      FRVS.ParaStyles[Result].Standard := False;
      if Assigned(FRVE.OnAddStyle) then
        FRVE.OnAddStyle(FRVE, FRVS.ParaStyles[Result]);
    end;
    ParaInfo.free;
  end;

var i: integer;
  subnode: TSectionBase;
  CurrentPara: integer;
  FirstSection: Boolean;
  AnchorName: String;
begin
  AnchorName := GetBlockID(block, FTargetBlockList);
  if AnchorName<>'' then
    RVData.AddNamedCheckpointExTag(AnchorName, False, RVEMPTYTAG);
  CurrentPara := DoParaFromBlock;
  inc(Indent, FRVS.PixelsToUnits(block.Indent));
  FirstSection := True;
  for i := 0 to block.DrawList.Count - 1 do
  begin
    subnode := block.DrawList.Items[i];
    if subnode is TTableBlock then
      ImportTableBlock(RVData, TTableBlock(subnode), CurrentPara)

    else if subnode is TTableAndCaptionBlock then
      ImportTableAndCaptionBlock(RVData,
        TTableAndCaptionBlock(subnode), CurrentPara)

    else if subnode is ThtmlTable then
      ImportTable(RVData, ThtmlTable(subnode), block, CurrentPara)

    else if subnode is TListBlock then
      ImportListBlock(RVData, TListBlock(subnode), CurrentPara, ListRec, Indent)

    else if subnode is TBlockLI then
      ImportBlockLi(RVData, TBlockLI(subnode), CurrentPara, ListRec, Indent)

    else if subnode is THRBlock then
      ImportHRBlock(RVData, THRBlock(subnode), CurrentPara)

    else if subnode is TBlock then
      ImportBlock(RVData, TBlock(subnode), CurrentPara, ListRec, Indent)

    else if subnode is TSection then
      ImportSection(RVData, TSection(subnode), CurrentPara, FirstSection)

    else if subnode is THorzLine then
      ImportHorzLine(RVData, THorzLine(subnode), CurrentPara)

    else
      LogNotImplemented('ImportABlock: ' + subnode.ClassName);

    FirstSection := not (subnode is TSection);
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVHTMLViewImporter.ImportListBlock(RVData: TCustomRVData; block: TListBlock;
  DefParaNo: Integer; var ListRec: TRVCurrentListRecord; Indent: TRVStyleLength);
var OldCounter: Integer;
begin
  inc(ListRec.Level);
  OldCounter := ListRec.Counter;
  ListRec.Counter := 0;
  if ListRec.Level=0 then begin
    ListRec.ListNo := -1;
    ListRec.AssignedLevels := [];
  end;
  ImportBlock(RVData, block, DefParaNo, ListRec, Indent);
  dec(ListRec.Level);
  ListRec.Counter := OldCounter;
end;
{------------------------------------------------------------------------------}
type PPositionRec = ^PositionRec;
function GetBackgroundPos(const PRec: PtPositionRec): PPositionRec;
begin
  {$IFDEF GOOGLECODE}
  Result := @(PRec.X);
  {$ELSE}
  Result := @(PRec[1]);
  {$ENDIF}
end;
function GetBackgroundPosX(const PRec: PtPositionRec): PPositionRec;
begin
  {$IFDEF GOOGLECODE}
  Result := @(PRec.X);
  {$ELSE}
  Result := @(PRec[1]);
  {$ENDIF}
end;
function GetBackgroundPosY(const PRec: PtPositionRec): PPositionRec;
begin
  {$IFDEF GOOGLECODE}
  Result := @(PRec.Y);
  {$ELSE}
  Result := @(PRec[2]);
  {$ENDIF}
end;
{------------------------------------------------------------------------------}
procedure TRVHTMLViewImporter.ImportTableCell(tablerow: TRVTableRow;
  cellobj: TCellObj; colindex: integer;
  var SumHPadding, SumVPadding, PaddingHCount, PaddingVCount: Integer;
  IsTableBorder: Boolean);
var Source: String;
var
  tablecell: TRVTableCellData;
  i, CurrentPara, BorderWidth: integer;
  FirstSection: Boolean;
  subnode: TSectionBase;
  gr: TGraphic;
  table: TRVTableItemInfo;
  BackgroundPos: PPositionRec;
  ListRec: TRVCurrentListRecord;
begin
  if colindex < tablerow.Count then
    tablecell := tablerow.Items[colindex]
  else
    tablecell := tablerow.Add;

  if (cellobj.ColSpan=0) or (cellobj.RowSpan=0) then
    exit;

  table := tablecell.GetTable;

  BorderWidth := FRVS.PixelsToUnits(GetCellBorderWidth(cellobj));
  {$IFDEF GOOGLECODE}
  table.CellBorderWidth := BorderWidth;
  {$ELSE}
  if (table.CellBorderWidth < BorderWidth) and not IsTableBorder then
    table.CellBorderWidth := BorderWidth;
  {$ENDIF}
  if table.CellBorderStyle in [rvtbRaised, rvtbLowered] then
    if TRVHtmlBorderStyle(cellobj.MargArray[BorderLeftStyle]) = bssSolid then
      table.CellBorderStyle := rvtbColor;

  AddCellHPadding(cellobj, SumHPadding, PaddingHCount);
  AddCellVPadding(cellobj, SumVPadding, PaddingVCount);

  if
    {$IFDEF GOOGLECODE}
    (TRVHtmlBorderStyle(cellobj.MargArray[BorderLeftStyle]) <> bssNone) or
    (TRVHtmlBorderStyle(cellobj.MargArray[BorderRightStyle]) <> bssNone) or
    (TRVHtmlBorderStyle(cellobj.MargArray[BorderTopStyle]) <> bssNone) or
    (TRVHtmlBorderStyle(cellobj.MargArray[BorderBottomStyle]) <> bssNone)
    {$ELSE}
    cellobj.BorderStyle<>bssNone
    {$ENDIF}
    then begin
    FillVisibleBorders(cellobj.MargArray, tablecell.VisibleBorders);
    if table.CellBorderStyle = rvtbColor then
      tablecell.BorderColor := GetBlockBorderColor(cellobj.MargArray)
    else if table.CellBorderStyle = rvtbRaisedColor then begin
      tablecell.BorderLightColor := GetBlockBorderLightColor(cellobj.MargArray);
      tablecell.BorderColor := GetBlockBorderDarkColor(cellobj.MargArray);
    end;
  end
  else if not IsTableBorder then
    tablecell.VisibleBorders.SetAll(False);

  if (cellobj.BGImage<>nil) and
    (cellobj.BGImage.Source<>'') then begin
    Source := cellobj.BGImage.Source;
    if (FBasePath<>'') and (Copy(Source, 1, 3)='Z:\') then
      Delete(Source, 1, 3);
    Source := GetFullFileName(Source);
    LoadPicture(FRVE.RVData, Source, gr, 0, 0, False);
    tablecell.BackgroundImage := gr;
    if gr<>nil then begin
      BackgroundPos := GetBackgroundPos(cellobj.PRec);
      if BackgroundPos.PosType={$IFDEF GOOGLECODE}bpDim{$ELSE}pDim{$ENDIF} then
        if BackgroundPos.RepeatD then
          tablecell.BackgroundStyle := rvbsTiled
        else
          tablecell.BackgroundStyle := rvbsStretched
      else
        tablecell.BackgroundStyle := rvbsCentered;
    end;
    gr.Free;
  end;

  tablecell.Clear;

  if cellobj.Cell.BkGnd then
    tablecell.Color := cellobj.Cell.BkColor;

  case cellobj.VAlign of
    ATop: tablecell.VAlign := rvcTop;
    AMiddle: tablecell.VAlign := rvcMiddle;
    ABottom: tablecell.VAlign := rvcBottom;
  end;

  {$IFDEF GOOGLECODE}
  if cellobj.SpecWd.VType=wtPercent then
    tablecell.BestWidth := -(cellobj.SpecWd.Value div 10)
  else if cellobj.SpecWd.VType=wtAbsolute then
    tablecell.BestWidth := FRVS.PixelsToUnits(cellobj.SpecWd.Value);
  if cellobj.SpecHt.VType=wtAbsolute then
    tablecell.BestHeight := FRVS.PixelsToUnits(cellobj.SpecHt.Value);
  {$ELSE}
  if cellobj.AsPercent then
    tablecell.BestWidth := -(cellobj.WidthAttr div 10)
  else
    tablecell.BestWidth := FRVS.PixelsToUnits(cellobj.WidthAttr);
  tablecell.BestHeight := FRVS.PixelsToUnits(cellobj.SpecHt);
  {$ENDIF}

  CurrentPara := 0;
  InitListRec(ListRec);
  FirstSection := True;
  for i := 0 to cellobj.Cell.Count - 1 do begin
    subnode := cellobj.Cell.Items[i];
    if subnode is TTableBlock then
      ImportTableBlock(TCustomRVDAta(tablecell),
        TTableBlock(subnode), CurrentPara)

    else if subnode is TTableAndCaptionBlock then
      ImportTableAndCaptionBlock(TCustomRVDAta(tablecell),
        TTableAndCaptionBlock(subnode), CurrentPara)

    else if subnode is ThtmlTable then
      ImportTable(TCustomRVDAta(tablecell),
        ThtmlTable(subnode), nil, CurrentPara)

    else if subnode is TBlockLI then
      ImportBlockLi(TCustomRVDAta(tablecell),
        TBlockLI(subnode), CurrentPara, ListRec, 0)

    else if subnode is THRBlock then
      ImportHRBlock(TCustomRVDAta(tablecell),
        THRBlock(subnode), CurrentPara)

    else if subnode is TListBlock then
      ImportListBlock(TCustomRVDAta(tablecell),
        TListBlock(subnode), CurrentPara, ListRec, 0)

    else if subnode is TBlock then
      ImportBlock(TCustomRVDAta(tablecell),
        TBlock(subnode), CurrentPara, ListRec, 0)

    else if subnode is TSection then
      ImportSection(TCustomRVDAta(tablecell),
        TSection(subnode), CurrentPara, FirstSection)

    else if subnode is THorzLine then
      ImportHorzLine(TCustomRVDAta(tablecell),
        THorzLine(subnode), CurrentPara)

    else
      LogNotImplemented('CheckTableCell: ' + subnode.ClassName);

    FirstSection := not (subnode is TSection);
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVHTMLViewImporter.ImportHorzLine(RVData: TCustomRVData;
  blockhl: THorzLine; ParaNo: Integer);
var
  st: TRVBreakStyle;
  Item: TRVBreakItemInfo;
begin
  case {$IFDEF GOOGLECODE}GetBlockBorderStyle(blockhl.OwnerBlock){$ELSE}THackHorzLine(blockhl).MyBlock.BorderStyle{$ENDIF} of
    bssDashed:
      st := rvbsDashed;
    bssDotted:
      st := rvbsDotted;
    else
      if (not blockhl.NoShade) and (blockhl.Color=clNone) then
        st := rvbs3d
      else
        st := rvbsLine;
  end;
  Item := TRVBreakItemInfo.CreateEx(RVData, blockhl.VSize, st,
    blockhl.Color);
  Item.SameAsPrev := False;
  Item.ParaNo := ParaNo;
  RVData.AddItem('', Item);
end;
{------------------------------------------------------------------------------}
procedure TRVHTMLViewImporter.ImportTable(RVData: TCustomRVData;
  htmltable: ThtmlTable; parentBlock: TBlock; ParaNo: Integer);
var
  max, i, j: integer;
  table: TRVTableItemInfo;
  gr: TGraphic;
  SumHPadding, SumVPadding, PaddingHCount, PaddingVCount, CellCount, CellPadding: Integer;
  IsTableBorder: Boolean;
  Source: String;
  BackgroundPos: PPositionRec;
begin
  table := TRVTableItemInfo.Create(RVData);
  table.Rows.Clear;
  table.Color := clNone;
  if htmltable.BkGnd then
    table.Color := htmltable.BkColor

  else if Assigned(parentBlock) and
    (parentBlock is TTableBlock) and
    TTableBlock(parentblock).BkGnd
    then
    table.Color := TTableBlock(parentblock).BkColor;

  table.ParaNo := ParaNo;
  {$IFDEF GOOGLECODE}
  table.BorderWidth := FRVS.PixelsToUnits(htmltable.BorderWidth);
  table.CellVSpacing := FRVS.PixelsToUnits(htmltable.CellSpacingVert);
  table.CellHSpacing := FRVS.PixelsToUnits(htmltable.CellSpacingHorz);
  table.BorderVSpacing := FRVS.PixelsToUnits(htmltable.CellSpacingVert);
  table.BorderHSpacing := FRVS.PixelsToUnits(htmltable.CellSpacingHorz);
  {$ELSE}
  table.BorderWidth := FRVS.PixelsToUnits(htmltable.Border);
  table.CellVSpacing := FRVS.PixelsToUnits(htmltable.CellSpacing);
  table.CellHSpacing := FRVS.PixelsToUnits(htmltable.CellSpacing);
  table.BorderVSpacing := FRVS.PixelsToUnits(htmltable.CellSpacing);
  table.BorderHSpacing := FRVS.PixelsToUnits(htmltable.CellSpacing);
  {$ENDIF}
  table.CellPadding := FRVS.PixelsToUnits(htmltable.CellPadding);
  CellPadding := htmltable.CellPadding;
  //table.Options := table.Options + [rvtoCellBelowBorders];
  if ((ParentBlock as TTableBlock).BGImage<>nil) and
    (TTableBlock(ParentBlock).BGImage.Source<>'') then begin
    Source := TTableBlock(ParentBlock).BGImage.Source;
    if (FBasePath<>'') and (Copy(Source, 1, 3)='Z:\') then
      Delete(Source, 1, 3);
    Source := GetFullFileName(Source);
    LoadPicture(RVData, Source, gr, 0, 0, False);
    table.BackgroundImage := gr;
    if gr<>nil then begin
      BackgroundPos := GetBackgroundPosX(TTableBlock(ParentBlock).PRec);
      if BackgroundPos.PosType={$IFDEF GOOGLECODE}bpDim{$ELSE}pDim{$ENDIF} then
        if BackgroundPos.RepeatD then
          table.BackgroundStyle := rvbsTiled
        else
          table.BackgroundStyle := rvbsStretched
      else
        table.BackgroundStyle := rvbsCentered;
    end;
    gr.Free;
  end;
  {$IFnDEF GOOGLECODE}
  IsTableBorder := (ParentBlock as TTableBlock).TableBorder;
  if IsTableBorder then begin
    if (htmltable.BorderColorLight = clBtnHighLight) and (htmltable.BorderColorDark = clBtnShadow) then
      table.BorderStyle := rvtbRaised
    else if (htmltable.BorderColorLight = clBtnShadow) and (htmltable.BorderColorDark = clBtnHighLight) then
      table.BorderStyle := rvtbLowered
    else if htmltable.BorderColorLight <> htmltable.BorderColorDark then begin
      table.BorderLightColor := htmltable.BorderColorLight;
      table.BorderColor := htmltable.BorderColorDark;
      table.BorderStyle := rvtbRaisedColor;
      end
    else begin
      table.BorderColor := htmltable.BorderColorDark;
      table.BorderStyle := rvtbColor;
    end;
    end
  else
  {$ELSE}
  IsTableBorder := False;
  {$ENDIF}
  begin
    table.BorderWidth := FRVS.PixelsToUnits(GetBlockBorderWidth(ParentBlock.MargArray));
    table.BorderStyle := rvtbColor;
    table.BorderLightColor := GetBlockBorderLightColor(ParentBlock.MargArray);
    table.BorderColor := GetBlockBorderDarkColor(ParentBlock.MargArray);
    if table.BorderLightColor <> table.BorderColor then
      if (table.BorderLightColor=clBtnHighlight) and (table.BorderColor=clBtnShadow) then
        table.BorderStyle := rvtbLowered
      else if (table.BorderColor=clBtnHighlight) and (table.BorderLightColor=clBtnShadow) then
        table.BorderStyle := rvtbRaised
      else
        table.BorderStyle := rvtbRaisedColor
    else
      table.BorderStyle := rvtbColor;
    if table.BorderWidth>0 then
      FillVisibleBorders(ParentBlock.MargArray, table.VisibleBorders);
  end;
  if (table.BorderWidth>0) and IsTableBorder then
    table.CellBorderWidth := FRVS.PixelsToUnits(1)
  else
    table.CellBorderWidth := 0;

  if table.BorderStyle in [rvtbRaised, rvtbLowered] then begin
    table.CellBorderStyle := rvtbLowered;
    table.CellBorderColor := table.BorderColor;
    table.CellBorderLightColor := table.BorderLightColor;
    end
  else begin
    table.CellBorderStyle := rvtbRaisedColor;
    table.CellBorderColor := table.BorderColor;
    table.CellBorderLightColor := table.BorderColor;
  end;
  SumHPadding := 0;
  SumVPadding := 0;
  PaddingHCount := 0;
  PaddingVCount := 0;
  CellCount   := 0;
  for i := 0 to htmltable.Rows.Count - 1 do
    ImportTableRow(table, TCellList(htmltable.Rows[i]), i,
    SumHPadding, SumVPadding, PaddingHCount, PaddingVCount, CellCount,
    IsTableBorder);
  if PaddingHCount>0 then
    table.CellHPadding := FRVS.PixelsToUnits(Round((SumHPadding + CellPadding*(CellCount*2-PaddingHCount))/(CellCount*2)));
  if PaddingVCount>0 then
    table.CellVPadding := FRVS.PixelsToUnits(Round((SumVPadding + CellPadding*(CellCount*2-PaddingVCount))/(CellCount*2)));
  max := 1;
  for i := 0 to table.Rows.Count - 1 do
    if max < table.Rows.Items[i].Count then
      max := table.Rows.Items[i].Count;
  for i := 0 to table.Rows.Count - 1 do
    while max > table.Rows.Items[i].Count do
      table.Rows.Items[i].Add;
  if table.Rows.Count = 0 then
    table.Rows.Add(1);
  for i := 0 to htmltable.Rows.Count - 1 do
    for j := 0 to TCellList(htmltable.Rows[i]).Count - 1 do
      DoMerge(table, i, j, TCellObj(TCellList(htmltable.Rows[i]).Items[j]).ColSpan,
        TCellObj(TCellList(htmltable.Rows[i]).Items[j]).RowSpan);
  table.BestWidth := 0;
  if (parentBlock <> nil) and (parentBlock is TTableBlock) then begin
    if TTableBlock(parentBlock).AsPercent then begin
      //if TTableBlock(parentBlock).WidthAttr <> 1000 then
        table.BestWidth := -Round(TTableBlock(parentBlock).WidthAttr / 10)
    end
    else
      table.BestWidth := FRVS.PixelsToUnits(TTableBlock(parentBlock).WidthAttr);
  end;
  RVData.AddItem('', table);
end;
{------------------------------------------------------------------------------}
procedure TRVHTMLViewImporter.ImportTableRow(table: TRVTableItemInfo;
  row: TCellList; rowindex: integer; var SumHPadding, SumVPadding,
  PaddingHCount, PaddingVCount, CellCount: Integer;
  IsTableBorder: Boolean);
var
  i: integer;
  tablerow: TRVTableRow;
begin
  if rowindex < table.Rows.Count then
    tablerow := table.Rows[rowindex]
  else
  begin
    tablerow := table.Rows.Add(row.Count);
    tablerow.Clear;
  end;
  for i := 0 to row.Count - 1 do begin
    ImportTableCell(tablerow, TCellObj(row.Items[i]), i,
      SumHPadding, SumVPadding, PaddingHCount, PaddingVCount,
      IsTableBorder);
    if (TCellObj(row.Items[i]).ColSpan>0) and (TCellObj(row.Items[i]).RowSpan>0) then
      inc(CellCount);
  end;
end;
{------------------------------------------------------------------------------}
type
  TImageObjHack = class (TImageObj)
  end;

procedure TRVHTMLViewImporter.ImportPicture(RVData: TCustomRVData;
  ImageObj: TImageObj; FO: TFontObj; const ParaNo: Integer);
var
  gr: TGraphic;
  GraphicItem: TRVGraphicItemInfo;
  w, h: Integer;
  VAlign: TRVVAlign;
  StyleNo: Integer;
  Tag: TRVTag;
  ItemName: TRVRawByteString;
  Source: String;
begin
  if (ImageObj.Source = '') then
   exit;
  Source := ImageObj.Source;
  if (FBasePath<>'') and (Copy(Source, 1, 3)='Z:\') then
    Delete(Source, 1, 3);
  Source := GetFullFileName(Source);
  w := TImageObjHack(ImageObj).SpecWidth;
  h := TImageObjHack(ImageObj).SpecHeight;
  LoadPicture(RVData, Source, gr, w, h, True);
  if gr <> nil then begin
    {$IFDEF GOOGLECODE}
    case THackFloatingObj(ImageObj).VertAlign of
      ATop:     VAlign := rvvaAbsTop;
      ABottom:  VAlign := rvvaAbsBottom;
      AMiddle:  VAlign := rvvaAbsMiddle;
      ANone:
        case THackFloatingObj(ImageObj).Floating of
          ALeft:  VAlign := rvvaLeft;
          ARight: VAlign := rvvaRight;
          else    VAlign := rvvaBaseline;
        end;
      ALeft:    VAlign := rvvaLeft;
      ARight:   VAlign := rvvaRight;
      else      VAlign := rvvaBaseline;
    end;
    {$ELSE}
    case THackFloatingObj(ImageObj).ObjAlign of
      ATop:     VAlign := rvvaAbsTop;
      ABottom:  VAlign := rvvaAbsBottom;
      AMiddle:  VAlign := rvvaAbsMiddle;
      ALeft:    VAlign := rvvaLeft;
      ARight:   VAlign := rvvaRight;
      else      VAlign := rvvaBaseline;
    end;
    {$ENDIF}
    StyleNo := rvsPicture;
    if ImgSrcToImageName then
      ItemName := TRVAnsiString(Source)
    else
      ItemName := '';
    Tag := RVEMPTYTAG;
    if FO.UrlTarget.URL <> '' then
      MakeTag(FO.UrlTarget.URL, StyleNo, Tag, ItemName);
    GraphicItem := CreateRichViewItem(StyleNo, RVData) as TRVGraphicItemInfo;
    GraphicItem.Tag := Tag;
    GraphicItem.Image := gr;
    GraphicItem.VAlign := VAlign;
    GraphicItem.Spacing := GetSpacing(imageobj);
    if (w > 0) then
      GraphicItem.ImageWidth := FRVS.PixelsToUnits(w);
    if (h > 0) then
      GraphicItem.ImageHeight := FRVS.PixelsToUnits(h);
    GraphicItem.OuterHSpacing :=
      Round((FRVS.PixelsToUnits(THackFloatingObj(ImageObj).HSpaceL) +
        FRVS.PixelsToUnits(THackFloatingObj(ImageObj).HSpaceR)) / 2);
    GraphicItem.OuterVSpacing :=
      Round((FRVS.PixelsToUnits(THackFloatingObj(ImageObj).VSpaceT) +
        FRVS.PixelsToUnits(THackFloatingObj(ImageObj).VSpaceB)) / 2);
    GraphicItem.ParaNo := ParaNo;
    RVData.AddItem(Source, GraphicItem);
    RVData.SetItemExtraStrProperty(RVData.ItemCount-1, rvespImageFileName,
      RemoveLineBreaks(Source));
    RVData.SetItemExtraStrProperty(RVData.ItemCount-1, rvespAlt,
      imageobj.Alt);
    if ImageObj.OwnerBlock <> nil then
    begin
      GraphicItem.BorderColor := GetBlockBorderColor(ImageObj.OwnerBlock.MargArray);
    end;
    {$IFDEF GOOGLECODE}
    if imageobj.Alt<>imageobj.Title then
      RVData.SetItemExtraStrProperty(RVData.ItemCount-1, rvespHint,
        imageobj.Title);
    {$ELSE}
    if imageobj.Alt<>THackFloatingObj(imageobj).ImageTitle then
      RVData.SetItemExtraStrProperty(RVData.ItemCount-1, rvespHint,
        THackFloatingObj(imageobj).ImageTitle);
    {$ENDIF}
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVHTMLViewImporter.ImportSection(RVData: TCustomRVData;
  sec: TSection; DefParaNo: Integer; FirstSection: Boolean);

  function DoRVTextStyleFromFontObject(FO: TFontObj): integer;
  var FontInfo: TFontInfo;
  {$IFNDEF RVDONOTUSESTYLETEMPLATES}
     StyleTemplateIndex: Integer;
  {$ENDIF}
  begin
    FontInfo := TFontInfo.Create(nil);
    FontInfo.Assign(FRVS.TextStyles.Items[0]);
    FontInfo.FontName := FO.TheFont.Name;
    FontInfo.Charset := FO.TheFont.Charset;
    FontInfo.Size := FO.TheFont.Size;
    FontInfo.Style := FO.TheFont.Style;
    FontInfo.Color := FO.TheFont.Color;
    FontInfo.BackColor := FO.TheFont.bgColor;
    FontInfo.Jump := FO.URL <> '';
    if FontInfo.Jump then begin
      {$IFNDEF RVDONOTUSESTYLETEMPLATES}
      if RVData.UseStyleTemplates then begin
        if FontInfo.Jump then begin
          StyleTemplateIndex := FRVS.StyleTemplates.FindByName(RVHYPERLINKSTYLETEMPLATENAME);
          if StyleTemplateIndex>=0 then
            FontInfo.StyleTemplateId := FRVS.StyleTemplates[StyleTemplateIndex].Id
          else
            FontInfo.StyleTemplateId := -1;
          end
        else
          FontInfo.StyleTemplateId := -1;
      end;
     {$ENDIF}
      FontInfo.HoverColor := FO.FIArray.Ar[HLFont].iColor;
      FontInfo.HoverBackColor := FO.FIArray.Ar[HLFont].iBgColor;
      if FontInfo.HoverColor=FontInfo.Color then
        FontInfo.HoverColor := clNone;
      if FontInfo.HoverBackColor=FontInfo.BackColor then
        FontInfo.HoverBackColor := clNone;
      if (fsUnderline in FO.FIArray.Ar[HLFont].iStyle) and
         not (fsUnderline in FontInfo.Style) then
        FontInfo.HoverEffects := FontInfo.HoverEffects+[rvheUnderline];
    end;
    FontInfo.Unicode := FRVS.TextStyles[0].Unicode;
    if FontInfo.Unicode and (FontInfo.Charset<>SYMBOL_CHARSET) then
      FontInfo.Charset := DEFAULT_CHARSET;
    case FO.SScript of
      ASub: FontInfo.SubSuperScriptType := rvsssSubscript;
      ASuper: FontInfo.SubSuperScriptType := rvsssSuperScript;
    end;
    if FO.SScript in [ASub, ASuper] then
      FontInfo.Size := RV_GetDefSubSuperScriptSizeRev(FontInfo.Size);
    result := FRVS.TextStyles.FindSuchStyle(0, FontInfo,
      RVAllFontInfoProperties);
    if Result < 0 then
    begin
      Result := FRVS.TextStyles.Add.Index;
      FRVS.TextStyles[Result].Assign(FontInfo);
      FRVS.TextStyles[Result].Standard := False;
      if Assigned(FRVE.OnAddStyle) then
        FRVE.OnAddStyle(FRVE, FRVS.TextStyles[Result]);
    end;
    FontInfo.free;
  end;

  {
  function GetMaxHeight: Integer;
  var
    I: Integer;
  begin
    Result := 0;
    for I := 0 to sec.Fonts.Count - 1 do
      if TFontObj(sec.Fonts.Items[I]).TheFont.tmHeight > Result then
        Result := TFontObj(sec.Fonts.Items[I]).TheFont.tmHeight;
    end;
   }

  function DoParaFromSection: integer;
  var
    ParaInfo: TParaInfo;
    Alignment: TRVAlignment;
    LineSpacing: Integer;
  begin
    Alignment := GetRVAlignment(sec.Justify);
    //if (sec.LineHeight<0) or (sec.Fonts.Count=0) then
      LineSpacing := 100
    //else
    //  LineSpacing := Round(sec.LineHeight*100/Abs(GetMaxHeight))
    ;
    if (DefParaNo>=0) and
      (FRVS.ParaStyles[DefParaNo].Alignment=Alignment) and
      // ATTENTION!!! If the compiler shows an error on FirstLineIndent,
      // open HTMLSubs.pas and move TSection's FirstLineIndent from PRIVATE to PUBLIC
      (FRVS.ParaStyles[DefParaNo].FirstIndent=FRVS.PixelsToUnits(sec.FirstLineIndent)) and
      (FRVS.ParaStyles[DefParaNo].LineSpacing=LineSpacing) then begin
      Result := DefParaNo;
      exit;
    end;
    ParaInfo := TParaInfo.Create(nil);
    if DefParaNo>=0 then
      ParaInfo.Assign(FRVS.ParaStyles.Items[DefParaNo]);
    ParaInfo.Alignment := Alignment;
    ParaInfo.FirstIndent := FRVS.PixelsToUnits(sec.FirstLineIndent);
    ParaInfo.LineSpacing := LineSpacing;
    ParaInfo.LineSpacingType := rvlsPercent;
    Result := FRVS.ParaStyles.FindSuchStyle(0, ParaInfo,
      RVAllParaInfoProperties);
    if Result < 0 then begin
      Result := FRVS.ParaStyles.Add.Index;
      FRVS.ParaStyles.Items[Result].Assign(ParaInfo);
      FRVS.ParaStyles.Items[Result].Standard := False;
      {$IFDEF RVFULL}
      if Assigned(FRVE.OnAddStyle) then
        FRVE.OnAddStyle(FRVE, ParaInfo);
      {$ENDIF}
    end;
    ParaInfo.free;
  end;

  procedure ImportText(FO: TFontObj; s: WideString; ParaNo: integer);
  var Tag: TRVTag;
    StyleNo: integer;
    ItemName: TRVRawByteString;
  begin
    StyleNo := DoRVTextStyleFromFontObject(FO);
    if s = #8 then
      s := '';
    Tag := RVEMPTYTAG;
    ItemName := '';
    if FO.UrlTarget.URL <> '' then
      MakeTag(FO.UrlTarget.URL, StyleNo, Tag, ItemName);
    if not FirstSection then
      RVData.SetAddParagraphMode(False);
    RVData.AddNLWTag(s, StyleNo, ParaNo, Tag);
    if not FirstSection then
      RVData.SetAddParagraphMode(True);
  end;

var
  {$IFnDEF GOOGLECODE}Index,{$ENDIF}
  imagecount, controlscount: integer;
  Start: PWideChar;
  FO, FOLast: TFontObj;
  {$IFDEF GOOGLECODE}
  ImgObj, CtrlObj: TFloatingObj;
  {$ELSE}
  ImgObj: TFloatingObj;
  CtrlObj: TFormControlObj;
  {$ENDIF}
  s: WideString;
  cont: integer;
  CurrentPara: integer;
begin
  if sec.AnchorName and (FAnchorIndex<FAnchorList.Count) then begin
    RVData.AddNamedCheckpointExTag(TDocPosItem(FAnchorList.Items[FAnchorIndex]).ID,
      False, RVEMPTYTAG);
    inc(FAnchorIndex);
  end;
  // ATTENTION!!! If the compiler shows an error on Buff,
  // open HTMLSubs.pas and move TSection's Buff from PRIVATE to PUBLIC
  if Length(sec.Buff)<=0 then
    exit;
  CurrentPara := DoParaFromSection;
  Start := sec.Buff;
  FOLast := sec.Fonts.GetFontObjAt(0{$IFnDEF GOOGLECODE}, Index{$ENDIF});
  DoRVTextStyleFromFontObject(FOLast);
  cont := Length(sec.Buff);
  if sec.Buff[cont-1]=' ' then
    dec(cont);
  while cont > 0 do begin
    FO := sec.Fonts.GetFontObjAt(Start - sec.Buff {$IFnDEF GOOGLECODE}, Index{$ENDIF});
    {$IFDEF GOOGLECODE}
    imagecount := sec.Images.GetObjectAt(Start - sec.Buff, ImgObj);
    controlscount := sec.FormControls.GetObjectAt(Start - sec.Buff, CtrlObj);
    {$ELSE}
    imagecount := sec.Images.GetImageCountAt(Start - sec.Buff);
    controlscount := TFormControlList(sec.FormControls).GetControlCountAt(Start - sec.Buff);
    {$ENDIF}
    if controlscount = 0 then begin
      {$IFnDEF GOOGLECODE}
      CtrlObj := TFormControlList(sec.FormControls).FindControl(Start - sec.Buff);
      {$ENDIF}
      if CtrlObj is TFormControlObj then
        s := s + '[' + TFormControlObj(CtrlObj).Value + ']';
      end
    else if imagecount = 0 then begin
      if s <> '' then begin
        ImportText(FOlast, s, CurrentPara);
        CurrentPara := -1;
        s := '';
      end;
      {$IFnDEF GOOGLECODE}
      ImgObj := sec.Images.FindImage(Start - sec.Buff);
      {$ENDIF}
      if ImgObj is TImageObj then begin
        ImportPicture(RVData, TImageObj(ImgObj), FO, CurrentPara);
        CurrentPara := -1;
      end;
      end
    else begin
      if FO = FOlast then
        s := s + Start[0]
      else begin
        ImportText(FOlast, s, CurrentPara);
        CurrentPara := -1;
        s := Start[0];
        FOlast := FO;
      end;
    end;
    dec(cont);
    inc(Start, 1);
  end;
  if s <> '' then
    ImportText(FOlast, s, CurrentPara)
end;
{------------------------------------------------------------------------------}
type
  TRVListLevelHack = class (TRVListLevel)
  end;


procedure TRVHTMLViewImporter.ImportBlockLi(RVData: TCustomRVData;
  blockli: TBlockLI; ParaNo: Integer; var ListRec: TRVCurrentListRecord;
  Indent: TRVStyleLength);

  procedure DoListFromBlock(sec: THackBlockLi; var ResetNumbering: Boolean);
  var
    ListStyle, OriginalListStyle: TRVListInfo;
    ll: TRVListLevel;
    i: Integer;
  begin
    ResetNumbering := False;
    if ListRec.ListNo>=0 then begin
      OriginalListStyle := RVData.GetRVStyle.ListStyles[ListRec.ListNo];
      if ListRec.Level in ListRec.AssignedLevels then begin
        ListStyle := TRVListInfo.Create(nil);  // a list that can be modified
        ListStyle.Assign(OriginalListStyle);
        end
      else begin
        ListStyle := OriginalListStyle;  // a list that can be modified
      end;
      end
    else begin
      OriginalListStyle := nil;
      ListStyle := TRVListInfo.Create(nil); // completely new list
    end;
    for i := ListStyle.Levels.Count to ListRec.Level do
      ListStyle.Levels.Add;
    ll := ListStyle.Levels[ListRec.Level];
    ll.Font.Assign(sec.ListFont); // <- if the compiler shows error here, read
                                       // the comments at the beginning of this file
    ll.FormatString := '%'+IntToStr(ListRec.Level)+':s.';
    // ll.StartFrom := sec.ListNumb;
    case sec.ListStyleType of
      lbDecimal:    ll.ListType := rvlstDecimal;
      lbLowerAlpha: ll.ListType := rvlstLowerAlpha;
      lbUpperAlpha: ll.ListType := rvlstUpperAlpha;
      lbLowerRoman: ll.ListType := rvlstLowerRoman;
      lbUpperRoman: ll.ListType := rvlstUpperRoman;
      lbDisc: begin
          ll.ListType := rvlstBullet;
          ll.FormatString := #183;
          ll.Font.Name := 'Symbol';
          ll.Font.Charset := SYMBOL_CHARSET;
        end;
      lbCircle: begin
          ll.ListType := rvlstBullet;
          ll.FormatString := 'o';
          ll.Font.Name := 'Courier New';
        end;
      lbSquare: begin
          ll.ListType := rvlstBullet;
          ll.FormatString := #$A7;
          ll.Font.Name := 'Wingdings';
          ll.Font.Charset := SYMBOL_CHARSET;
        end;
      lbNone, lbBlank: begin
          ll.ListType := rvlstBullet;
          ll.FormatString := '';
        end;
    end;
    ll.MarkerIndent := Indent-FRVS.PixelsToUnits(24);
    if ll.MarkerIndent<0 then
      ll.MarkerIndent := 0;
    ll.FirstIndent := 0;
    ll.LeftIndent := Indent;
    {
    ll.MarkerIndent := sec.ContentLeft-24;  //TODO width of marker
    if ll.MarkerIndent<0 then
      ll.MarkerIndent := 0;
    ll.FirstIndent := 0;
    ll.LeftIndent := sec.ContentLeft;
    }
    if ListStyle=OriginalListStyle then begin
      Include(ListRec.AssignedLevels, ListRec.Level);
      inc(ListRec.Counter);
      ResetNumbering := (sec.ListNumb<>ListRec.Counter) and ll.HasNumbering;
      exit; // we have modified directly in TRVStyle
    end;
    if (OriginalListStyle=nil) or
      ((ListRec.Level in ListRec.AssignedLevels) and
       not TRVListLevelHack(OriginalListStyle.Levels[ListRec.Level]).IsSimpleEqual(ll)) then begin
      RVData.GetRVStyle.ListStyles.Add.Assign(ListStyle);
      ListRec.ListNo := RVData.GetRVStyle.ListStyles.Count-1;
      ListRec.Counter := sec.ListNumb;
      ResetNumbering := (sec.ListNumb<>1) and ll.HasNumbering;
      FRVS.ListStyles[ListRec.ListNo].Assign(ListStyle);
      FRVS.ListStyles[ListRec.ListNo].Standard := False;
      {$IFDEF RVFULL}
      if Assigned(FRVE.OnAddStyle) then
        FRVE.OnAddStyle(FRVE, ListStyle);
      {$ENDIF}
      end
    else begin
      inc(ListRec.Counter);
      ResetNumbering := (sec.ListNumb<>ListRec.Counter) and ll.HasNumbering;
    end;
    Include(ListRec.AssignedLevels, ListRec.Level);
    ListStyle.Free;
  end;

var
  ItemCount: integer;
  ResetNumbering: Boolean;
begin
  if ListRec.Level<0 then begin
    ImportBlock(RVData, TBlock(blockli), ParaNo, ListRec, Indent);
    exit;
  end;
  DoListFromBlock(THackBlockLi(blockli), ResetNumbering);
  ItemCount := RVData.ItemCount;
  ImportBlock(RVData, TBlock(blockli), ParaNo, ListRec, Indent);
  if RVData.ItemCount > ItemCount then begin
    RVData.SetListMarkerInfo(ItemCount, ListRec.ListNo, ListRec.Level,
      THackBlockLi(blockli).ListNumb, ParaNo, ResetNumbering);
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVHTMLViewImporter.ImportBody(RVData: TCustomRVData;
  bodyblock: TBodyBlock);
var ListRec: TRVCurrentListRecord;
begin
  InitListRec(ListRec);
  ImportBlock(RVData, TBlock(bodyblock), 0, ListRec, 0);
end;
{------------------------------------------------------------------------------}
procedure TRVHTMLViewImporter.ImportHRBlock(RVData: TCustomRVData;
  blockhr: THRBlock; DefParaNo: Integer);
var ListRec: TRVCurrentListRecord;
begin
  InitListRec(ListRec);
  ImportBlock(RVData, TBlock(blockhr), DefParaNo, ListRec, 0);
end;
{------------------------------------------------------------------------------}
procedure TRVHTMLViewImporter.ImportTableAndCaptionBlock(RVData: TCustomRVData;
  tableandcaption: TTableAndCaptionBlock; ParaNo: Integer);
var ListRec: TRVCurrentListRecord;
begin
  InitListRec(ListRec);
  ImportBlock(RVData, TBlock(tableandcaption), ParaNo, ListRec, 0);
end;
{------------------------------------------------------------------------------}
procedure TRVHTMLViewImporter.ImportTableBlock(RVData: TCustomRVData;
  tableblock: TTableBlock; ParaNo: Integer);
var ListRec: TRVCurrentListRecord;
begin
  InitListRec(ListRec);
  ImportBlock(RVData, TBlock(tableblock), ParaNo, ListRec, 0);
end;
{------------------------------------------------------------------------------}
constructor TRVHTMLViewImporter.Create(AOwner: TComponent);
begin
  inherited;
  AutoHyperlinks := True;
  UseCommonBulletLists := True;
  CompleteLists := True;
  DefaultHMargin := 5;
  DefaultVMargin := 5;
  {
  FDefFontName := 'Times New Roman';
  FDefPreFontName := 'Courier New';
  FDefFontSize := 12;
  FDefFontColor := clWindowText;
  FDefLinkColor := clBlue;
  FDefColor     := clWindow;
  }
end;

{
function TRVHTMLViewImporter.LoadFromFile(const FileName: String;
  rve: TCustomRichView): Boolean;
var Stream: TFileStream;
begin
  Result := False;
  try
    Stream := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
    try
      LoadFromStream(Stream, rve);
    finally
      Stream.Free;
    end;
    Result := True;
  except
  end;
end;

procedure TRVHTMLViewImporter.LoadFromStream(Stream: TStream;
  rve: TCustomRichView);
var s: String;
begin
  SetLength(s, Stream.Size-Stream.Position);
  Stream.ReadBuffer(PChar(s)^, Length(s));
  LoadFromString(s, rve);
end;

procedure TRVHTMLViewImporter.LoadFromString(const s: String;
  rve: TCustomRichView);
var SectionList: TSectionList;
begin
  SectionList := TSectionList.Create(rve, rve);
  try
    SectionList.SetFonts(FDefFontName, DefPreFontName, FDefFontSize, FDefFontColor,
      FDefLinkColor, FDefLinkColor, FDefLinkColor, FDefColor, False, FUndelineLinks,
      DEFAULT_CHARSET, rve.LeftMargin, rve.TopMargin);
    ParseHTMLString(s, SectionList, nil, nil, nil, nil);
    DoImport(SectionList, rve);
  finally
    SectionList.Free;
  end;
end;
}
{------------------------------------------------------------------------------}
const AUTOMARGIN = -12348765;

procedure TRVHTMLViewImporter.DoImport(SectionList: {$IFDEF GOOGLECODE}ThtDocument{$ELSE}TSectionList{$ENDIF};
  rve: TCustomRichView; AppendMode: Boolean);
var i: integer;
  sec: TSectionBase;
  gr: TGraphic;
  bmp: TBitmap;
  Color: TColor;
  DocPos: TDocPosItem;
  BlockID: TBlockIDItem;
  BackgroundPosX,BackgroundPosY: PPositionRec;
begin
  FRVE := rve;
  FRVS := rve.Style;
  FListCount := FRVS.ListStyles.Count;
  if not AppendMode then begin
    FRVE.BackgroundStyle := bsNoBitmap;
    FRVE.Clear;
  end;
  FAnchorList := TRVList.Create;
  FTargetBlockList  := TRVList.Create;
  for i := 0 to SectionList.IDNameList.Count-1 do
    if SectionList.IDNameList.Objects[i] is TChPosObj then begin
      DocPos := TDocPosItem.Create;
      DocPos.ID := SectionList.IDNameList.Strings[i];
      // if the compiler shows error on the next line, read the comments
      // at the beginning of this file
      DocPos.Pos := THackChPosObj(SectionList.IDNameList.Objects[i]).ChPos;
      FAnchorList.Add(DocPos);
      end
    else if SectionList.IDNameList.Objects[i] is TBlock then begin
      BlockID := TBlockIDItem.Create;
      BlockID.ID := SectionList.IDNameList.Strings[i];
      BlockID.Block := SectionList.IDNameList.Objects[i];
      FTargetBlockList.Add(BlockID);
    end;
  FAnchorList.Sort(DocPosSortCompare);
  FAnchorIndex := 0;
  if not AppendMode then begin
    Color := clWindow;
    if SectionList.BackGround <> clNone then
      Color := SectionList.BackGround;
      if FRVE.Color=clNone then
        FRVE.Style.Color := Color
      else
        FRVE.Color := Color;
    {$IFDEF GOOGLECODE}
    if SectionList.BackgroundImage <> nil then
    {$ELSE}
    if SectionList.BackgroundBitmap <> nil then
    {$ENDIF}
    begin
      BackgroundPosX := GetBackgroundPosX(
        SectionList.{$IFDEF GOOGLECODE}BackgroundImagePosition{$ELSE}BackgroundPRec{$ENDIF});
      BackgroundPosY := GetBackgroundPosY(
        SectionList.{$IFDEF GOOGLECODE}BackgroundImagePosition{$ELSE}BackgroundPRec{$ENDIF});
      if (BackgroundPosX.PosType in [{$IFDEF GOOGLECODE}bpCenter, bpPercent{$ELSE}pCenter, pPercent{$ENDIF}]) or
         (BackgroundPosY.PosType in [{$IFDEF GOOGLECODE}bpCenter, bpPercent{$ELSE}pCenter, pPercent{$ENDIF}]) then
          FRVE.BackgroundStyle := bsCentered
      else if (BackgroundPosX.PosType in [{$IFDEF GOOGLECODE}bpDim{$ELSE}pDim{$ENDIF}]) or
         (BackgroundPosY.PosType in [{$IFDEF GOOGLECODE}bpDim{$ELSE}pDim{$ENDIF}]) then begin
        if BackgroundPosX.RepeatD or BackgroundPosY.RepeatD then
            if BackgroundPosX.Fixed or BackgroundPosY.Fixed then
              FRVE.BackgroundStyle := bsTiled
            else
              FRVE.BackgroundStyle := bsTiledAndScrolled
          else
            FRVE.BackgroundStyle := bsStretched;
        end
      else begin
        case BackgroundPosX.PosType of
          {$IFDEF GOOGLECODE}bpTop, bpLeft, bpCenter{$ELSE}pTop, pLeft, pCenter{$ENDIF}:
            case BackgroundPosY.PosType of
              {$IFDEF GOOGLECODE}bpTop, bpLeft, bpCenter{$ELSE}pTop, pLeft, pCenter{$ENDIF}:
                FRVE.BackgroundStyle := bsTopLeft;
              else
                FRVE.BackgroundStyle := bsBottomLeft;
            end;
          else
            case BackgroundPosY.PosType of
              {$IFDEF GOOGLECODE}bpTop, bpLeft, bpCenter{$ELSE}pTop, pLeft, pCenter{$ENDIF}:
                FRVE.BackgroundStyle := bsTopRight;
              else
                FRVE.BackgroundStyle := bsBottomRight;
            end;
        end;
      end;
      LoadPicture(FRVE.RVData,
        GetFullFileName(
          SectionList.{$IFDEF GOOGLECODE}BackgroundImageName{$ELSE}BitmapName{$ENDIF}),
        gr, 0, 0, False);
      if gr <> nil then begin
        try
          FRVE.BackgroundBitmap.Assign(gr);
        except
          bmp := TBitmap.Create;
          try
            bmp.Width := gr.Width;
            bmp.Height := gr.Height;
            bmp.Canvas.Draw(0, 0, gr);
            FRVE.BackgroundBitmap.Assign(bmp);
          finally
            bmp.Free;
          end;
        end;
        gr.Free;
      end;
    end;
  end;

  for i := 0 to SectionList.Count - 1 do
  begin
    sec := SectionList.Items[i];
    if sec is TBodyBlock then
    begin
      if not AppendMode then begin
        FRVE.LeftMargin := TBodyBlock(sec).MargArray[MarginLeft];
        if FRVE.LeftMargin=AUTOMARGIN then
          FRVE.LeftMargin := DefaultHMargin;
        FRVE.RightMargin := TBodyBlock(sec).MargArray[MarginRight];
        if FRVE.RightMargin=AUTOMARGIN then
          FRVE.RightMargin := DefaultHMargin;
        FRVE.TopMargin := TBodyBlock(sec).MargArray[MarginTop];
        if FRVE.TopMargin=AUTOMARGIN then
          FRVE.TopMargin := DefaultVMargin;
        FRVE.BottomMargin := TBodyBlock(sec).MargArray[MarginBottom];
        if FRVE.BottomMargin=AUTOMARGIN then
          FRVE.BottomMargin := DefaultVMargin;
      end;
      ImportBody(FRVE.RVData, TBodyBlock(sec));
    end;
  end;
  FAnchorList.Free;
  FAnchorList := nil;
  FTargetBlockList.Free;
  FTargetBlockList := nil;
  if CompleteLists then
    for i := FListCount to FRVS.ListStyles.Count-1 do
      DoCompleteList(i);
end;
{------------------------------------------------------------------------------}
procedure TRVHTMLViewImporter.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if (Operation=opRemove) then begin
   if (AComponent=FHTMLViewer) then
     FHTMLViewer := nil;
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVHTMLViewImporter.DoCompleteList(ListNo: Integer);
var Indent: TRVStyleLength;
    i: Integer;
    Numbered: Boolean;
    ListStyle: TRVListInfo;
    NewLevel: TRVListLevel;

begin
  ListStyle := FRVS.ListStyles[ListNo];
  Numbered := ListStyle.AllNumbered;
  if ListStyle.Levels.Count=0 then
    Indent := 0
  else if ListStyle.Levels.Count=1 then
    Indent := ListStyle.Levels[0].LeftIndent
  else
    Indent := ListStyle.Levels[ListStyle.Levels.Count-1].LeftIndent-
      ListStyle.Levels[ListStyle.Levels.Count-2].LeftIndent;
  for i := ListStyle.Levels.Count to 8 do begin
    NewLevel := ListStyle.Levels.Add;
    NewLevel.Assign(ListStyle.Levels[i-1]);
    NewLevel.LeftIndent   := NewLevel.LeftIndent+Indent;
    NewLevel.MarkerIndent := NewLevel.MarkerIndent+Indent;
    NewLevel.FormatString := '%'+IntToStr(i)+':s.';
    if Numbered then
      NewLevel.ListType := rvlstDecimal
    else
      case i of
        1: begin
          NewLevel.ListType := rvlstBullet;
          NewLevel.FormatString := 'o';
          NewLevel.Font.Name := 'Courier New';
        end;
      else
        begin
          NewLevel.ListType := rvlstBullet;
          NewLevel.FormatString := #$A7;
          NewLevel.Font.Name := 'Wingdings';
          NewLevel.Font.Charset := SYMBOL_CHARSET;
        end;
      end;
  end;
end;

{==============================================================================}
procedure Register;
begin
  RegisterComponents('RichView', [TRVHTMLViewImporter]);
end;


end.